<?php

use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\Unicode;
use Drupal\Component\Utility\Xss;
use Drupal\Core\Link;
use Drupal\Core\Logger\RfcLogLevel;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Page callback: Displays a listing of database log messages.
 *
 * Messages are truncated at 56 chars. Full-length messages can be viewed on the
 * message details page.
 *
 * @ingroup logging_severity_levels
 */
function notice_killer_overview() {
  $rows = [];
  $classes = notice_killer_severity_class_map();

  $header = [
    '', // Icon column.
    [
      'data' => t('Last logged'),
      'field' => 'w.last_timestamp',
      'sort' => 'desc',
    ],
    ['data' => t('First logged'), 'field' => 'w.wid', 'sort' => 'desc'],
    ['data' => t('Appearances'), 'field' => 'w.count', 'sort' => 'desc'],
    t('Message'),
    ['data' => t('User'), 'field' => 'u.name'],
  ];

  $query = \Drupal::database()
    ->select('notice_killer', 'w')
    ->extend('PagerDefault')
    ->extend('TableSort');
  $query->leftJoin('users', 'u', 'w.uid = u.uid');
  $query->fields('w', [
      'wid',
      'uid',
      'severity',
      'timestamp',
      'message',
      'variables',
      'last_timestamp',
      'count',
      'type',
    ])->addField('u', 'name');
  $result = $query->limit(50)->orderByHeader($header)->execute();

  foreach ($result as $dblog) {
    $rows[] = [
      'data' => [
        // Cells.
        ['class' => 'icon'],
        \Drupal::service('date.formatter')
          ->format($dblog->last_timestamp, 'custom', 'd/M/Y'),
        \Drupal::service('date.formatter')
          ->format($dblog->timestamp, 'custom', 'd/M/Y'),
        $dblog->count,
        [
          '#theme' => 'notice_killer_message',
          '#event' => $dblog,
          '#link' => TRUE,
        ],
        [
          '#theme' => 'username',
          '#account' => $dblog->uid,
        ],
      ],
      // Attributes for tr.
      'class' => [
        Html::getClass('dblog-' . $dblog->type),
        $classes[$dblog->severity],
      ],
    ];
  }

  $build['dblog_table'] = [
    '#theme' => 'table',
    '#header' => $header,
    '#rows' => $rows,
    '#attributes' => ['id' => 'admin-dblog'],
    '#empty' => t('No log messages available.'),
  ];
  $build['dblog_pager'] = ['#theme' => 'pager'];

  return $build;
}

/**
 * @return string[]
 */
function notice_killer_severity_class_map(): array {
  return [
    RfcLogLevel::DEBUG => 'dblog-debug',
    RfcLogLevel::INFO => 'dblog-info',
    RfcLogLevel::NOTICE => 'dblog-notice',
    RfcLogLevel::WARNING => 'dblog-warning',
    RfcLogLevel::ERROR => 'dblog-error',
    RfcLogLevel::CRITICAL => 'dblog-critical',
    RfcLogLevel::ALERT => 'dblog-alert',
    RfcLogLevel::EMERGENCY => 'dblog-emergency',
  ];
}

/**
 * Returns HTML for a log message.
 *
 * @param array $variables
 *   An associative array containing:
 *   - event: An object with at least the message and variables properties.
 *   - link: (optional) Format message as link, event->wid is required.
 *
 * @ingroup themeable
 */
function theme_notice_killer_message($variables) {
  $output = '';
  $event = $variables['event'];
  // Check for required properties.
  if (isset($event->message) && isset($event->variables)) {
    // Messages without variables or user specified text.
    if ($event->variables === 'N;') {
      $output = $event->message;
    }
    // Message to translate with injected variables.
    else {
      $output = t($event->message, unserialize($event->variables));
    }
    if ($variables['link'] && isset($event->wid)) {
      // Truncate message to 80 chars.
      $output = Unicode::truncate(Xss::filter($output, []), 80, TRUE, TRUE);
      $url = Url::fromRoute('notice_killer.event', ['wid' => $event->wid], ['html' => TRUE]);
      $output = new Link($output, $url);
    }
  }
  return $output;
}

/**
 * Page callback: Displays details about a specific database log message.
 *
 * @param int $id
 *   Unique ID of the database log message.
 *
 * @return array|string
 *   If the ID is located in the Database Logging table, a build array in the
 *   format expected by drupal_render(); otherwise, an empty string.
 *
 * @see dblog_menu()
 */
function notice_killer_event($id) {
  $severity = notice_killer_severity_levels();
  $result = \Drupal::database()
    ->query('SELECT w.*, u.name, u.uid FROM {notice_killer} w INNER JOIN {users} u ON w.uid = u.uid WHERE w.wid = :id', [':id' => $id])
    ->fetchObject();
  if ($dblog = $result) {
    $rows = [
      [
        ['data' => t('Type'), 'header' => TRUE],
        t($dblog->type),
      ],
      [
        ['data' => t('First appearance date'), 'header' => TRUE],
        date_format($dblog->timestamp, 'long'),
      ],
      [
        ['data' => t('Last appearance date'), 'header' => TRUE],
        date_format($dblog->last_timestamp, 'long'),
      ],
      [
        ['data' => t('Appearances'), 'header' => TRUE],
        $dblog->count,
      ],
      [
        ['data' => t('User'), 'header' => TRUE],
        [
          '#theme' => 'username',
          '#account' => $dblog->uid,
        ]
      ],
      [
        ['data' => t('Location'), 'header' => TRUE],
        new Link($dblog->location, Url::fromUri($dblog->location)),
      ],
      [
        ['data' => t('Referrer'), 'header' => TRUE],
        new Link($dblog->referer, Url::fromUri($dblog->referer)),
      ],
      [
        ['data' => t('Message'), 'header' => TRUE],
        [
          '#theme' => 'notice_killer_message',
          '#event' => $dblog,
        ],
      ],
      [
        ['data' => t('Severity'), 'header' => TRUE],
        $severity[$dblog->severity],
      ],
      [
        ['data' => t('Hostname'), 'header' => TRUE],
        Html::escape($dblog->hostname),
      ],
    ];
    $build['dblog_table'] = [
      '#theme' => 'table',
      '#rows' => $rows,
      '#attributes' => ['class' => ['dblog-event']],
    ];

    // Add in the backtrace if we have one.
    if (!empty($dblog->backtrace) && ($backtrace = unserialize($dblog->backtrace))) {
      // Print the stacktrace.
      $build['backtrace'] = [
        '#type' => 'fieldset',
        '#title' => t('Backtrace'),
      ];

      $backtrace_header = [
        t('Function'),
        t('File'),
        t('Line'),
      ];
      $backtrace_rows = [];
      foreach ($backtrace as $row) {
        $backtrace_rows[] = [
          Html::escape(notice_killer_backtrace_function_name($row)),
          isset($row['file']) ? Html::escape($row['file']) : '',
          isset($row['line']) ? Html::escape($row['line']) : '',
        ];
      }

      $build['backtrace']['table'] = [
        '#theme' => 'table',
        '#rows' => $backtrace_rows,
        '#header' => $backtrace_header,
      ];

      if (function_exists('kprint_r')) {
        $build['backtrace']['kprint'] = [
          '#markup' => kprint_r($backtrace, TRUE),
        ];
      }
      else {
        // @TODO Confirm.
        $build['backtrace']['kprint'] = [
          '#markup' => t('You will get more information about the backtrace if you enable the <a href="@devel_link">Devel module</a>.', ['@devel_link' => url('https://www.drupal.org/project/devel', ['external' => TRUE])]),
        ];
      }
    }

    // Add in some actions.
    $build['actions']['delete'] = [
      '#type' => 'link',
      '#href' => 'admin/reports/notices/event/' . $id . '/delete/' . drupal_get_token('notice_killer_' . $id),
      '#title' => t('Delete this log entry'),
      '#options' => [
        'attributes' => [
          'class' => [
            'button',
          ],
        ],
      ],
    ];

    return $build;
  }
  else {
    return '';
  }
}

/**
 * Return data on the function name for a specific backtrace row.
 *
 * @param $backtrace_row
 *   The row from the backtrace to print.
 *
 * @return string
 *   The computed function name.
 */
function notice_killer_backtrace_function_name($backtrace_row) {
  $function = isset($backtrace_row['function']) ? Html::escape($backtrace_row['function']) : '';
  if (isset($backtrace_row['class']) && isset($backtrace_row['type'])) {
    $function = $backtrace_row['class'] . $backtrace_row['type'] . $function;
  }
  return $function;
}
