<?php

namespace Drupal\notice_killer\Logger;

use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\DatabaseExceptionWrapper;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Logger\LogMessageParserInterface;
use Drupal\Core\Logger\RfcLoggerTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\dblog\Logger\DbLog;
use Psr\Log\LoggerInterface;

class NoticeKillerLogger implements LoggerInterface {
  use RfcLoggerTrait;
  use DependencySerializationTrait;

  /**
   * The database connection object.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The message's placeholders parser.
   *
   * @var \Drupal\Core\Logger\LogMessageParserInterface
   */
  protected $parser;

  /**
   * @var \Drupal\Core\Config\ConfigFactory
   */
  protected $configFactory;

  /**
   * Standard dbLog logger as fallback.
   *
   * @var \Drupal\dblog\Logger\dbLog
   */
  protected $dbLog;

  /**
   * Constructs a DbLog object.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection object.
   * @param \Drupal\Core\Logger\LogMessageParserInterface $parser
   *   The parser to use when extracting message variables.
   */
  public function __construct(Connection $connection, LogMessageParserInterface $parser, ConfigFactory $configFactory) {
    $this->connection = $connection;
    $this->parser = $parser;
    $this->configFactory = $configFactory;
    $this->dblog = new DbLog($this->connection, $this->parser);

    $this->prevent_dblog = (bool) $this->configFactory->get('notice_killer.settings')->get('notice_killer_prevent_dblog_notices');
  }

  /**
   * Log the notice to our table.
   *
   * @param int $level
   *   Severity level.
   * @param string $message
   *   Message for the notice.
   * @param array $context
   *   Context/variables to go with the message.
   */
  public function log($level, $message, array $context = []) {
    // We only want to interfere with php notices.
    if ($context['channel'] != 'php') {
      $this->dbLog->log($level, $message, $context);
    }
    else {
      // Check whether we want to dblog this notice or not.
      if (!$this->prevent_dblog) {
        $this->dbLog->log($level, $message, $context);
      }
      else {
        $this->nkLog($context, $message);
      }
    }
  }

  /**
   * @param array $context
   * @param string $message
   *
   * @return void
   * @throws \Exception
   */
  public function nkLog(array $context, string $message): void {
    if (notice_killer_logged_message($context)) {
      $message_args = [
        '%file' => $context['%file'],
        '%function' => $context['%function'],
        '%type' => $context['%type'],
        '@backtrace_string' => $context['@backtrace_string'],
        '@message' => $context['@message']->__toString(),
      ];
      $config = $this->configFactory->get('notice_killer.settings');
      $message_args['%line'] = $config->get('notice_killer_ignore_line_numbers') ? $context['%line'] : 'N/A';
      if ($config->get('notice_killer_ignore_file_location') && !empty($message_args['%file'])) {
        if (strpos($message_args['%file'], DRUPAL_ROOT) === 0) {
          $message_args['%file'] = substr($message_args['%file'], strlen(DRUPAL_ROOT));
        }
      }
      $message_markup = new TranslatableMarkup($message, $message_args);
      $message_string = $message_markup->render();
      $hash = sha1($message_string);

      // Log the message.
      $update = $this->connection->update('notice_killer')
        ->expression('count', 'count + 1')
        ->fields(['last_timestamp' => $context['timestamp']])
        ->condition('hash', $hash);
      $num_updated = $update->execute();
      if ($num_updated == 0) {
        // This is a new error we've not seen before.
        try {
          $context['hash'] = $hash;
          $context['wid'] = $this->connection->insert('notice_killer')->fields([
              'uid' => $context['uid'],
              'type' => substr($context['channel'], 0, 64),
              'message' => $message,
              'variables' => serialize($message_args),
              'severity' => $context['severity_level'],
              'link' => substr($context['link'], 0, 255),
              'location' => $context['request_uri'],
              'referer' => $context['referer'],
              'hostname' => substr($context['ip'], 0, 128),
              'timestamp' => $context['timestamp'],
              'hash' => $context['hash'],
              'last_timestamp' => $context['timestamp'],
              'backtrace' => serialize(notice_killer_backtrace_clean(notice_killer_get_backtrace(), $config->get('notice_killer_backtrace'))),
            ])->execute();

          notice_killer_notify_new_log($context);
        }
        catch (DatabaseExceptionWrapper $e) {
          // So it's possible that someone else has just inserted the notice, try the increment again.
          $update = $this->connection->update('notice_killer')
            ->expression('count', 'count + 1')
            ->fields(['last_timestamp' => $context['timestamp']])
            ->condition('hash', $hash);
          $update->execute();
        }
      }
    }
  }

}
